// const { default: G6 } = require("@antv/g6");
/**
 * Requires: utils.js 
 */

/**
 * Helper class for processing node or edge states.
 */
class _StateItems {
    /**
     * Constructs a helper object. 
     * @param {String} namespace Namespace for AntV G6 state machanism. 
     *                           Each state in AntV G6 is named as `namespace:stateName`.
     */
    constructor(namespace) {
        this.dest = {};
        this.namespace = namespace;
    }

    /**
     * Adds a state as an object with the following attributes:
     *   * `category`: Category of the state
     *   * `name`: Name of the state
     *   * `fullName`: Full name used for AntV G6, `fullName = 'namespace:name'`
     *   * `model`: Model of node style parameters, including `size`, `style`
     * 
     * Each state item added can be accessed via `states[name]` where `states` is the object
     * obtained by `get()` method. 
     * 
     * Supports method chaining. 
     * @param {String} category See above
     * @param {String} name See above
     * @param {Object} model See above
     * @returns this
     */
    add(category, name, model) {
        this.dest[name] = Object.freeze({
            'category': category,
            'name': name,
            'fullName': `${this.namespace}:${name}`,
            'model': model 
        });
        return this; 
    }

    /**
     * Gets the state collection object.
     * @returns A read-only object that contains all the states as attributes.
     */
    get() {
        return Object.freeze(this.dest); 
    }
}

/**
 * Gets the coordinate by X axis of given node.
 * @param {Object} node The node object
 * @returns A floating point, X axis coordinate.
 */
function _getX(node) {
    // A hack is used by accessing _cfg directly.
    return node._cfg.model.x;
}

/**
 * Gets the coordinate by Y axis of given node.
 * @param {Object} node The node object
 * @returns A floating point, Y axis coordinate.
 */
function _getY(node) {
    // A hack is used by accessing _cfg directly.
    return node._cfg.model.y; 
}

/**
 * Gets the coordinate by the specified axis of given node.
 * @param {Object} node The node object
 * @param {String} axis `x` or `y`
 * @returns A floating point, coordinate of the corresponding axis.
 */
function _getPos(node, axis) {
    return node._cfg.model[axis];
}

/**
 * Gets the coordinate of given node as an array of numbers `[x, y]`.
 * @param {Object} node The node object
 * @returns An array of length 2, `[x, y]`
 */
function _getPos2D(node) {
    return [_getX(node), _getY(node)];
}

/**
 * Gets the identifier of animation node of given indices.
 * 
 * Animation nodes are aggregated as G groups. Each group contains K nodes moving with different speed. 
 * @param {Number} i Group index, which group is this animation node in, ranged as 0, 1 ... G-1
 * @param {Number} j Node index in the group, ranged as 0, 1 ... K-1
 * @returns Identifier of the animation node.
 */
function _getAnimationNodeId(i, j) {
    return `@Animation(${i}, ${j})`;
}

/**
 * Checks whether the node identifier belongs to an animation node.
 * @param {String} id Identifier of the node.
 * @returns Whether the identifier is of an animation node.
 */
function _isAnimationNodeId(id) {
    // Consistent with _getAnimationNodeId
    return id.startsWith('@Animation');
}

/**
 * Gets the opacity of the given node or edge. 
 * @param {Object} obj The node or edge object.
 * @returns The opacity value
 */
function _getOpacity(obj) {
    return obj._cfg.model.style.opacity; 
}

/**
 * Creates a table of node states.
 * The states consist of the following categories:
 *   * `default`: contains state sub-names `default` and `boosted`;
 *   * `Ca`: contains state sub-names `default`, `source`, `boosted` and `animation`;
 *   * `Ca+`: Same as above;
 *   * `Cr`: Same as above; 
 *   * `Cr-`: Same as above.
 * Each state is named as the format `category,subName` (e.g. `Ca+,boosted`)
 * @param {Object} styles Arguments of node styles obtained via `readConfig().nodeStyles`
 * @param {String} namespace Which namespace is used in AntV G6 state mechanism, `nodeState` by default.
 * @returns The object of node state table.
 * @see _StateItems for details
 */
function createNodeStates(styles, namespace = 'nodeState') {
    const items = new _StateItems(namespace);
    ['default', 'boosted'].forEach((t) => {
        items.add('default', t, styles[t]); 
    });
    ['Ca', 'Ca+', 'Cr', 'Cr-'].forEach((s) => {
        ['default', 'boosted', 'source', 'animation'].forEach((t) => {
            items.add(s, `${s},${t}`, styles[s][t]); 
        });
    });
    return items.get(); 
}

/**
 * Creates a table of edge states.
 * @param {Object} styles Arguments of node styles obtained via `readConfig().edgeStyles`
 * @param {String} namespace Which namespace is used in AntV G6 state mechanism, `edgeState` by default.
 * @returns The object of edge state table.
 * @see _StateItems for details.
 */
function createEdgeStates(styles, namespace = 'edgeState') {
    const items =  new _StateItems(namespace);
    ['default', 'Ca', 'Ca+', 'Cr', 'Cr-'].forEach((t) => {
        items.add(t, t, styles[t]);
    });
    return items.get(); 
}

/**
 * Appends attributes to the canvas object and each node and edge inside the canvas.
 * This is a helper function that shall be called by `createCanvas()` function, 
 * after adding all the nodes (including the animation nodes) and edges.
 * 
 * The following attributed are added to the canvas object:
 *   * `n`: Number of nodes in the graph;
 *   * `m`: Number of edges in the graph; 
 *   * `propData`: Propagation data transformed from the `data` object;
 *   * `nodeStates`: Collection of node states;
 *   * `edgeStates`: Collection of edge states;
 *   * `nNodes`: An object of node count, with the following attributes (consistent of node state categories):
 *      * `default`: Number of nodes without any state, i.e. no message received yet;
 *      * `Ca`: Number of nodes of state Ca, i.e. with non-boosted positive message;
 *      * `Ca+`: Number of nodes of state Ca+, i.e. with boosted positive message;
 *      * `Cr`: Number of nodes of state Cr, i.e. with non-boosted negative message;
 *      * `Cr-`: Number of nodes of state Cr-, i.e. with boosted negative message. 
 *   * `nAnimationGroupsUsed`: How many groups of animation nodes are used currently;
 *   * `animationGroups`: Groups of animation nodes;
 *   * `staticNodes`: List of all the non-animation nodes, i.e. nodes originally in the graph.
 * 
 * `propData` contains the attributes given in `createEventObject` function (in utils.js). 
 * 
 * Each object in `sourceEvents` 2D-array contains the following attributes:
 *   * `node`: A reference to the node object in the canvas as information source;
 *   * `state`: A reference to the state object that the current node shall be.
 * 
 * Each object in `propagationEvents` 2D-array contains the following attributes:
 *   * `source`: A reference to the node object in the canvas which sends its information;
 *   * `target`: A reference to the node object in the canvas which receives the information;
 *   * `edge`: A reference to the corresponding edge object in the canvas;
 *   * `edgeState`: A reference to the edge state object that the edge shall be;
 *   * `animationState`: A reference to the node state object that the animation nodes along the edge shall be;
 *   * `targetState`: A reference to the node state object that the target node shall be. 
 * 
 * `animationGroups` is a list with length `G = propData.maxNPropagations`.
 * Each group `animationGroups[i]` (i = 0...G-1) is a list of length `args.nAnimationNodes`,
 * references to the animation nodes in the canvas, each assigned with different speed.
 * 
 * Each node object in the canvas has the following attributes appended:
 *   * `id`: Identifier of the node;
 *   * `state`: State of the node, one of the objects in `canvas.nodeStates`;
 *   * `isBoosted`: Whether this node is a boosted node;
 *   * (Animation nodes only) `speed`: Speed of the animation node. 
 * 
 * Each edge object in the canvas has the following attributes appended:
 *   * `id`: Identifier of the edge;
 *   * `state`: State of the edge, one of the objects in `canvas.edgeStates`.
 * 
 * @param {Object} canvas The canvas object obtained by calling `createCanvas()` function
 * @param {Object} graph The graph object obtained via `readGraph()` function
 * @param {Object} data The data object obtained via `readData()` function
 * @param {Object} args Additional arguments obtained via `readConfig()` function
 * @param {Object} nodeStates Node state collection obtained via `createNodeStates()` function
 * @param {Object} edgeStates Edge state collection obtained via `createEdgeStates()` function
 * @see createEventObject in input.js for details of attributes in `canvas.propData`.
 */
function _appendCanvasAttributes(canvas, graph, data, args, nodeStates, edgeStates) {
    // Appends attributes for the canvas 
    // (1) data (details see createEventObject() in input.js)
    createAttribute(canvas, 'propData', createEventObject(
        data.label, data.nRounds, data.boostedNodes, data.maxNPropagations
    ));
    // (2) Collection of node & edge states
    createAttribute(canvas, 'nodeStates', nodeStates);
    createAttribute(canvas, 'edgeStates', edgeStates);
    // (3) Object of n, m and counting variables
    createAttribute(canvas, 'n', graph.n);
    createAttribute(canvas, 'm', graph.m); 
    createAttribute(canvas, 'nNodes', {
        'default': graph.n, 'Ca': 0, 'Ca+': 0, 'Cr': 0, 'Cr-': 0
    });
    createAttribute(canvas, 'nAnimationGroupsUsed', 0);
    // (4) Groups animation nodes as `G = maxNPropagation` groups, each group has `K = nAnimationNodes` nodes.
    createAttribute(canvas, 'animationGroups', []);
    d3.range(data.maxNPropagations).forEach((i) => {
        canvas.animationGroups[i] = [];
        d3.range(args.nAnimationNodes).forEach((j) => {
            const node = canvas.findById(_getAnimationNodeId(i, j));
            // Puts to i-th group
            canvas.animationGroups[i].push(node);
            // Additional parameters for current animation node
            // Speed of animation nodes in a group are v, v+d, v+2d ... v+(K-1)d resp.
            node.speed = args.initialSpeed + j * args.speedIncrement
        });
    });

    // Appends attributes for each node
    canvas.getNodes().forEach((node) => {
        // (n-1) id: A hack is used here by accessing _cfg directly
        createAttribute(node, 'id', node._cfg.id);
        // (n-2) state: State of this node, one of attributes in canvas.nodeStates
        createAttribute(node, 'state', canvas.nodeStates.default);
    });
    // (n-3) isBoosted: true/false for boosted/non-boosted nodes 
    data.boostedNodes.forEach((id) => {
        createAttribute(canvas.findById(id), 'isBoosted', true); 
    });
    // False for others
    canvas.getNodes().forEach((node) => {
        if (node.isBoosted == undefined) {
            node.isBoosted = false; 
        }
    })

    // Appends attributes for each edge
    canvas.getEdges().forEach((edge) => {
        // (e-1) id: A hack is used here by accessing _cfg directly
        edge.id = edge._cfg.id;
        // (e-2) state: State of this edge, one of the attributes in canvas.edgeStates
        edge.state = canvas.edgeStates.default; 
    });

    // (Cont.) Appends attributes for the canvas
    // (5) List of static nodes
    createAttribute(canvas, 'staticNodes', canvas.findAll('node', (node) => {
        return !_isAnimationNodeId(node.id);
    }));

    // (1) cont. 
    d3.range(data.nRounds).forEach((t) => {
        // Transforms sourceEvents[t][]
        data.sourceEvents[t].forEach((rawEvt) => {
            const node = canvas.findById(rawEvt.id);
            canvas.propData.sourceEvents[t].push({
                'node': node,
                // Takes the source state
                'state': canvas.nodeStates[`${rawEvt.type},source`]
            });
        });
        // Transforms propagationEvents[t][]
        data.propagationEvents[t].forEach((rawEvt) => {
            const u = canvas.findById(rawEvt.source);
            const v = canvas.findById(rawEvt.target);
            canvas.propData.propagationEvents[t].push({
                'source': u,
                'target': v,
                'edge': canvas.findById(getEdgeId(u.id, v.id)),
                'edgeState': canvas.edgeStates[rawEvt.edgeType],
                // Animation node states are consistent of edge state (typically, the same color as the edge)
                'animationState': canvas.nodeStates[`${rawEvt.edgeType},animation`],
                // If v is a boosted node, then changes to the `'boosted'` alternative. Otherwise `'default'`.
                'targetState': canvas.nodeStates[`${rawEvt.targetType},${v.isBoosted ? 'boosted' : 'default'}`]
            });
        })
    });
}

/**
 * Creates am AntV G6 canvas object and initializes with given graph and propagation data.
 * @param {*} graph The graph object obtained via `readGraph()` function
 * @param {*} data  The data object obtained via `readData()` function
 * @param {string} container Label of the container element
 * @param {number} width Width of the canvas 
 * @param {number} height Height of the canvas
 * @param {*} args Additional arguments obtained via `readConfig()` function
 * @returns The canvas object created.
 */
function createCanvas(graph, data, container, width, height, args) {
    function extractArgs(attrName, defaultValue, mapFn = null) {
        let res = null; 
        if (args[attrName] == undefined) {
            console.log(`'${attrName}' uses default value: `, defaultValue);
            res = defaultValue;
        } else {
            res = args[attrName];
        }
        if (mapFn != null) {
            res = mapFn(res);
            console.log(`'${attrName}' maps to: `, res);
        }
        return res; 
    }

    const nodeStates = createNodeStates(args.nodeStyles);
    const edgeStates = createEdgeStates(args.edgeStyles);

    // Transforms node/edge states to [node|edge]StateStyles parameter: 
    //  namespace:state => model.style
    function createStateStylesArg(states) {
        const res = {};
        Object.values(states).forEach((obj) => {
            res[obj.fullName] = obj.model.style;
        });
        return res;
    }

    const fitViewPadding = extractArgs("fitViewPadding", 0, (r) => {
        // Maps from ratio to size value
        if (typeof(r) == "number") {
            return [r * height, r * width, r * height, r * width];
        } else {
            return [r[0] * height, r[1] * width, r[2] * height, r[3] * width];
        }
    });

    const canvas = new G6.Graph({
        container: container,
        width: width,
        height: height,
        // Default style of nodes
        defaultNode: args.nodeStyles.default,
        // Default style of edges
        defaultEdge: args.edgeStyles.default,
        // Let the graph fit the canvas, 
        fitView: true,
        fitViewPadding: fitViewPadding,
        // Node styles for each state (attribute 'size' etc. shall be set separately)
        nodeStateStyles: createStateStylesArg(nodeStates),
        // Graph layout style
        layout: extractArgs("graphLayout", {
            "type": "random",
            // Prevents overlapping of nodes
            "preventOverlap": true,
        }),
        animate: true,
        animateCfg: {
            duration: 1000,
        },
    });
    // Initializes with the graph contents
    canvas.read(graph);

    canvas.updateLayout({
        "type": "gForce",
        // Prevents overlapping of nodes
        "preventOverlap": true,
        "nodeStrength": 20,  // Default: 1000
        "edgeStrength": 200, // Default: 200,
        "onLayoutEnd": () => {
            console.log(`animate = ${canvas.get('animate')}`);
            console.log(`animateCfg = `, canvas.get('animateCfg'));
            canvas.fitView(fitViewPadding);
        }
    });
    
    // Adds animation nodes
    d3.range(data.maxNPropagations).forEach((i) => {
        d3.range(args.nAnimationNodes).forEach((j) => {
            canvas.addItem('node', {
                'id': _getAnimationNodeId(i, j)
            });
        });
    });
    canvas.refresh();
    // Appends attributes.
    _appendCanvasAttributes(canvas, graph, data, args, nodeStates, edgeStates);
    return canvas;
}

/**
 * Changes the state of the given node, with the following actions:
 *   * The new state is rendered to canvas immediately;
 *   * Count of nodes are updated;
 *   * The attribute `node.state` is updated as the current one.
 * 
 * If `node` is provided as a string, the node object is searched by `canvas.findById(node)`.
 * 
 * If `state` is provided as a string, the state is accessed via `canvas.nodeStates[state]`.
 * @param {Object} canvas The canvas object;
 * @param {Object | String} node The node object, or the identifier of the node;
 * @param {Object | String} state The state object, or the name of the state. 
 */
function setNodeState(canvas, node, state) {
    // Transforms state to Object
    if (typeof(state) == 'string' || state instanceof String) {
        state = canvas.nodeStates[state];
    }
    // Transforms node to Object inside canvas
    if (typeof(node) == 'string' || node instanceof String) {
        node = canvas.findById(node);
    }
    
    function incCount(state, delta = 1) {
        canvas.nNodes[state.category] += delta; 
    }

    // No nothing if no change occurs
    if (node.state == state) {
        return;
    } 
    // Otherwise, node states changes its category.
    // Decreases count of old state by 1
    incCount(node.state, -1);
    // Sets to new state, and then adds count of new state by 1
    node.state = state; 
    incCount(node.state, 1);
    // Sets style
    canvas.setItemState(node, 'nodeState', state.name);
    // Sets size
    canvas.updateItem(node, {size: state.model.size});
}

/**
 * Changes the state of given edge, with the following actions:
 *   * The new state is rendered to canvas immediately;
 *   * The attribute `node.state` is updated as the current one. 
 * 
 * If `edge` is provided as a string, the edge is searched by `canvas.findById(edge)`.
 * 
 * If `state` is provided as a string, the state object is accessed via `canvas.edgeStates`.
 * @param {Object} canvas The canvas object;
 * @param {Object | String} edge The edge object, or the identifier of the edge;
 * @param {Object | String} state The state object, or the name of the state. 
 */
function setEdgeState(canvas, edge, state) {
    // Transforms state to Object
    if (typeof(state) == 'string' || state instanceof String) {
        state = canvas.edgeStates[state]; 
    }
    // Transforms node to Object inside canvas
    if (typeof(edge) == 'string' || edge instanceof String) {
        edge = canvas.findById(node);
    }
    edge.state = state; 
    // Sets style
    canvas.updateItem(edge, {
        'style': state.model.style 
    });
}

/**
 * Initializes the first `g` groups of animation nodes in the canvas object, with the following actions:
 *   * Hides all the animation nodes. 
 * 
 * @param {Object} canvas The canvas object;
 * @param {Number} nGroups `g` above, How many groups of animation nodes to initialize. 
 *                         Initializes all the groups if `nGroups` not provided. 
 */
function initializeAnimationNodes(canvas, nGroups = null) {
    if (nGroups == null) {
        nGroups = canvas.animationGroups.length; // Takes all the groups by default. 
    }
    // Clears animation nodes
    d3.range(nGroups).forEach((i) => {
        canvas.animationGroups[i].forEach((node) => {
            canvas.hideItem(node); // Hides all the animation nodes requested before.
        })
    });
}

/**
 * Initializes the canvas at the beginning of each round, with the following actions:
 *   * Resets all the animation nodes requested, including:
 *     * Hides all the animation nodes requested;
 *     * Resets count of used animation node groups to 0.
 * 
 * @param {Object} canvas The canvas object. 
 * @see initializeAnimationNodes for details of resetting each animation node group. 
 */
 function initializeRound(canvas) {
    // Clears animation nodes
    initializeAnimationNodes(canvas, canvas.nAnimationGroupsUsed);
    // Resets group count
    canvas.nAnimationGroupsUsed = 0; 
}

/**
 * Initializes the canvas before starting the whole visualization process, with the following actions:
 *   * Sets each node state and edge state to default;
 *   * Resets all the animation nodes as hidden; 
 *   * Resets all the counting variables in canvas.nNodes.
 * 
 * @param {Object} canvas The canvas object.
 */
function initialize(canvas) {
    // Resets states of static nodes and edges 
    canvas.staticNodes.forEach((node) => {
        // For boosted nodes, sets to the `'boosted`' alternative. Otherwise `'default'`.
        setNodeState(canvas, node, node.isBoosted ? canvas.nodeStates.boosted : canvas.nodeStates.default);
    });
    canvas.getEdges().forEach((edge) => {
        setEdgeState(canvas, edge, canvas.edgeStates.default);
    })
    // Resets all the animation nodes 
    initializeAnimationNodes(canvas);
    canvas.nAnimationGroupsUsed = 0; 

    function checkAndReset(obj, attr, value = 0) {
        if (obj[attr] != value) {
            console.warn(`WARNING: ${attr} = ${obj[attr]} (expected: ${value}) after resetting node states. ` + 
            `This inconsistency infers that bugs may exist in node state changing process. `);
            obj[attr] = value;
        }
    }
    // Resets counting variables
    for (k in canvas.nNodes) {
        checkAndReset(canvas.nNodes, k, k === "default" ? canvas.n : 0);
    }
}

/**
 * Dispatches a group of animation nodes for visualization of the propagation event. 
 * 
 * Each animation node is updated as the following:
 *   * Sets node size by `state.model.size`;
 *   * Sets node style by `state.model.style`.
 * 
 * Besides, each animation node object in the requested group has the following attributes appended:
 *   * `sourcePosition`: An array of length 2 as `[x1, y1]` where (x1, y1) is the coordinate of the source position;
 *   * `targetPosition`: `[x2, y2]` where (x2, y2) is of the target position.
 * 
 * Position of the requested nodes are not changed, which is typically initialized before as outside the canvas 
 * via `initialize()` or `initializeRound()` functions.
 * 
 * @param {Object} canvas The canvas object;
 * @param {Object | String} state The state object or its name;
 * @param {Object} evt The propagation event object. 
 */
function requestAnimationGroup(canvas, state, evt) {
    // Transforms to state object. 
    if (typeof(state) == 'string' || state instanceof String) {
        state = canvas.nodeStates[state]; 
    }

    const group = canvas.animationGroups[canvas.nAnimationGroupsUsed];
    canvas.nAnimationGroupsUsed += 1;
    
    group.forEach((node) => {
        node.sourceNode = evt.source;
        node.targetNode = evt.target; 
        
        const sourcePosition = _getPos2D(evt.source);
        // Changes size and style. 
        canvas.updateItem(node, {
            'x': sourcePosition[0], // Initializes the position the same as the source node 
            'y': sourcePosition[1],
            'size': state.model.size,
            'style': state.model.style
        });
        // Shows the current animation node group on the canvas
        canvas.showItem(node); 
    });
}

/**
 * Displays all the events in the t-th round. 
 * 
 * The event lists are accessed via `canvas.propData.sourceEvents[t]` 
 * and `canvas.propData.propagationEvents[t]` resp.
 * 
 * State of propagation target node is delayed to display: 
 * its change is triggered when an animation node reaches from the corresponding source. 
 * 
 * The target node objects have the following attributes appended:
 *   * `delayedState`: a state object inside `canvas.nodeStates`, the state to be displayed later. 
 * 
 * @param {Object} canvas The canvas object;
 * @param {Number} t The index of round.
 */
function displayEvents(canvas, t) {
    // Source events
    canvas.propData.sourceEvents[t].forEach((evt) => {
        setNodeState(canvas, evt.node, evt.state);
    });
    // Propagation events
    canvas.propData.propagationEvents[t].forEach((evt) => {
        // Sets target node state: delaying machanism is applied.
        // Target state is changed only after animation node reaches the target. 
        evt.target.delayedState = evt.targetState;
        // Sets edge state according to source nodes 
        setEdgeState(canvas, evt.edge, evt.edgeState);
        // Triggers an animation group along current edge. 
        requestAnimationGroup(canvas, evt.animationState, evt);
    });
}

/**
 * Performs a frame of animation by moving all the animation nodes requested. 
 * 
 * Each animation node will be moved linearly from source to target position by progreess `min(1.0, node.speed * ratio)`.
 * Each node needs `1.0 / node.speed` of a round's duration to reach the destination,
 * and current time point is `ratio * 100%` time elapsed of current round. 
 * 
 * @param {Object} canvas The canvas object;
 * @param {Number} ratio A ratio value in the range [0, 1].
 * @returns Whether some nodes arrives the target position. 
 */
function doAnimation(canvas, ratio) {
    // No animation nodes activated at all
    if (canvas.nAnimationGroupsUsed == 0 || canvas.animationGroups.length == 0) {
        return false; 
    }

    let res = false; 
    // Each group moves separately
    d3.range(canvas.nAnimationGroupsUsed).forEach((i) => {
        const group = canvas.animationGroups[i]; 

        group.forEach((node) => {
            const lambda = Math.min(1.0, ratio * node.speed);
            const u = _getPos2D(node.sourceNode);
            const v = _getPos2D(node.targetNode);
            // Changes position: u + lambda * (v - u)
            canvas.updateItem(node, {
                'x': (1.0 - lambda) * u[0] + lambda * v[0],
                'y': (1.0 - lambda) * u[1] + lambda * v[1]
            });
            res |= (lambda == 1.0);
        });
    });

    return res; 
}

/**
 * Fades all the edges of positive / negative states.
 * @param {Object} canvas The canvas object;
 * @param {Object} fadeArgs The parameters for edge fading, obtained as `args.edgeFading` 
 *                          where `args` is obtained via `readConfig()` function. 
 */
function fadeEdges(canvas, fadeArgs) {
    canvas.getEdges().forEach((edge) => {
        // No change to default edges 
        if (edge.state.name == 'default') {
            return; 
        }
        const curArgs = fadeArgs[edge.state.name];
        const curOpacity = _getOpacity(edge);
        // No change if current opacity already reaches the minimal
        if (curOpacity <= curArgs.finalOpacity) {
            return; 
        }
        // Decrements opacity by delta
        const nextOpacity = Math.max(curOpacity - curArgs.delta, curArgs.finalOpacity);
        // Refreshes on the canvas
        canvas.updateItem(edge, {
            'style': {
                'opacity': nextOpacity
            }
        });
    });
}

/**
 * Main function to start the visualization loop process. 
 * 
 * At the beginning of each round, the `checkResumableCallback()` function will be called
 * to check whether the display animation shall continue (returns `true`) or pause (returns `false`).
 * An always-true function is used if not provided.
 * 
 * After the first frame of each round, the `roundStartCallback()` function will be called
 * to handle some other data processing (e.g. plots number of nodes of each state to a stacked bar, etc.).
 * A no-op function is used if not provided. 
 * 
 * @param {Array<Object>} canvasList An array of canvases.
 * @param {Object} args Arguments of the visualization process, obtained via `readConfig()` function. 
 * @param {() => boolean} checkResumableCallback Callback function to check whether the process shall be continued or paused.
 * @param {() => void} roundStartCallback Callback function after the first frame of a round. 
 */
function runVisualization(canvasList, args, checkResumableCallback = null, roundStartCallback = null) {
    // For single canvas object, wraps as a list
    if (!(canvasList instanceof Array)) {
        canvasList = [canvasList];
    }
    // Always true by default
    if (checkResumableCallback == null) {
        checkResumableCallback = () => true; 
    }
    // No-op by default
    if (roundStartCallback == null) {
        roundStartCallback = () => {}; 
    }

    // Total number of rounds. Rounds are indexed as R = 0, 1 ... nRounds-1
    const nRounds = canvasList[0].propData.nRounds;
    // How many frames for each iteration. For the R-th round, frames are indexed as i = 0, 1 ... nFramesPerRound-1
    const nFramesPerRound = args.nFramesPerRound;
    // Duration of each frame by milliseconds
    const interval = 1000.0 / args.fps; 
    // Count of frames
    let frameId = -1; 
    // How many ramaining frames before the next continue/pause check via checkResumableCallback()
    let remainingNFrames = nFramesPerRound; 
    // A flag variable whether all delayed states of target nodes has been displayed. 
    let delayedTriggered = false; 
    // Start time 
    let roundStartTime = Date.now();

    setInterval(() => {
        if (remainingNFrames == 0) {
            // If remaining frames run out, try to acquire more animation frames ...
            if (checkResumableCallback()) {
                remainingNFrames += nFramesPerRound;
            } else {
                // ... or pause the process
                return; 
            }
        }

        frameId += 1;
        remainingNFrames -= 1; 

        // Round index R
        const round = Math.floor(frameId / nFramesPerRound) % nRounds;
        // Frame index i in current round
        const frame = frameId % nFramesPerRound;
        
        if (frame == 0) {
            // Resets start time of current round
            roundStartTime = Date.now();
            // Initializes globally for the first round
            if (round == 0) {
                canvasList.forEach((canvas) => initialize(canvas));
            }
            // For the first frame of each round, performs message propagation
            canvasList.forEach((canvas) => {
                initializeRound(canvas); 
                // Resets delay flag
                delayedTriggered = false; 
                // Performs information source & propagation in current round
                displayEvents(canvas, round);
            });
            // Performs callback function
            roundStartCallback(); 
        }

        let hasArrived = false; 
        // Performs node animation and edge animation (edge fading) by 1 frame
        canvasList.forEach((canvas) => {
            hasArrived |= doAnimation(canvas, frame / (nFramesPerRound - 1));
            fadeEdges(canvas, args.edgeFading); 
        });

        // Triggers delayed states
        if (hasArrived && !delayedTriggered) {
            delayedTriggered = true; 
            canvasList.forEach((canvas) => {
                canvas.staticNodes.forEach((node) => {
                    if (node.delayedState != undefined) {
                        // Delayed states are triggered here
                        setNodeState(canvas, node, node.delayedState);
                        // Removes the delayed state after triggered. 
                        node.delayedState = undefined; 
                    }
                });
            });
        }

        // Dumps real FPS to console
        if (frame == nFramesPerRound - 1) {
            const endTime = Date.now();
            const realFps = 1000.0 / (endTime - roundStartTime) * nFramesPerRound;
            console.log(`Real FPS = ${Math.round(realFps)}`);
        }

    }, interval);
}